info = {
    'name': 'PushYourLuck',
    'description': 'As the name suggest, Push Your Luck is an artificial version of a "push your luck" game like, for instance, Can\'t Stop.',
    'context': 'ippc2018',
    'tags': '',
    'viz': 'None'
}
