info = {
    'name': 'TrafficBLX_SimplePhases',
    'description': 'BLX/QTM traffic signal control model with a fixed phase progression consisting of 4 phases. The goal is to control traffic lights to minimize total travel time.',
    'context': '',
    'tags': '',
    'viz': 'TrafficViz.TrafficVisualizer'
}
