info = {
    'name': 'TowerOfHanoi',
    'description': 'The classic tower of Hanoi puzzle, where disks must be stacked onto a given rod.',
    'context': 'arcade',
    'tags': '',
    'viz': 'None'
}