info = {
    'name': 'MountainCar',
    'description': 'A simple continuous MDP for the classical mountain car control problem',
    'context': 'ippc2023',
    'tags': '',
    'viz': 'MountainCarViz.MountainCarVisualizer',
    'generator': 'MountainCarInstanceGenerator.MountainCarInstanceGenerator'
}