info = {
    'name': 'Reservoir',
    'description': 'Continuous action version of management of the water level in interconnected reservoirs',
    'context': 'ippc2023',
    'tags': '',
    'viz': 'ReservoirViz.ReservoirVisualizer',
    'generator': 'ReservoirInstanceGenerator.ReservoirInstanceGenerator'
}