info = {
    'name': 'PowerGen',
    'description': 'A continuous simple power generation problem loosely modeled on the problem of unit commitment',
    'context': 'ippc2023',
    'tags': '',
    'viz': 'PowerGenViz.PowerGenVisualizer',
    'generator': 'PowerGenInstanceGenerator.PowerGenInstanceGenerator'
}