info = {
    'name': 'PowerGen_Continuous',
    'description': 'A simple continuous version of the power generation problem, loosely modeled on the problem of unit commitment.',
    'context': '',
    'tags': '',
    'viz': 'PowerGenViz.PowerGenVisualizer'
}