info = {
    'name': 'Knapsack',
    'description': 'Items of random weight and value are drawn, the goal is to place them into knapsacks of limited total weight while maximizing total value of all items.',
    'context': 'or',
    'tags': '',
    'viz': 'None'
}