info = {
    'name': 'HVAC',
    'description': 'Multi-zone and multi-heater HVAC control problem',
    'context': 'ippc2023',
    'tags': '',
    'viz': 'HVACViz.HVACVisualizer',
    'generator': 'HVACInstanceGenerator.HVACInstanceGenerator'
}