info = {
    'name': 'BinPacking',
    'description': 'Items of random weight are drawn, the goal is to place them into bins while minimizing the number of bins used and the total weight of each bin is within limits.',
    'context': 'or',
    'tags': '',
    'viz': 'None'
}