info = {
    'name': 'TriangleTireworld_POMDP',
    'description': 'In short, this problem was intended to be difficult for determinization/replanning approaches since the highest probability path to the goal is longer than other lower probability (but still possible) paths to the goal. This is the pomdp version',
    'context': 'ippc2014',
    'tags': '',
    'viz': 'None'
}
