info = {
    'name': 'Blackjack',
    'description': 'A card is randomly drawn from a deck and added to the player hand, the goal is to stop with the highest value of the hand without going over 21.',
    'context': 'arcade',
    'tags': '',
    'viz': 'None'
}