info = {
    'name': 'CooperativeRecon',
    'description': 'In this domain, the planner controls one or more planetary rovers that examine objects of interest in order to detect life and take a picture of it.',
    'context': 'ippc2018',
    'tags': '',
    'viz': 'None'
}
