info = {
    'name': 'Reservoir_Continuous',
    'description': 'Continuous action version of management of the water level in interconnected reservoirs.',
    'context': '',
    'tags': '',
    'viz': 'ReservoirViz.ReservoirVisualizer'
}