info = {
    'name': 'MarsRover',
    'description': 'Multi Rover Navigation, where a group of agent needs to harvest mineral',
    'context': 'ippc2023',
    'tags': '',
    'viz': 'MarsRoverViz.MarsRoverVisualizer',
    'generator': 'MarsRoverInstanceGenerator.MarsRoverInstanceGenerator'
}