info = {
    'name': 'SkillTeaching_MDP',
    'description': 'In this domain, the agent is trying to teach a series of skills to a student through the use of hints and multiple choice questions.',
    'context': 'ippc2014',
    'tags': '',
    'viz': 'None'
}
