info = {
    'name': 'Intruders_Continuous',
    'description': 'Continuous intruder detection problem on a unit square.',
    'context': '',
    'tags': '',
    'viz': 'IntrudersViz.IntrudersVisualizer'
}