info = {
    'name': 'UAV',
    'description': 'Continuous action space version of multi-UAV problem where a group of UAVs have to reach goal positions in the 3d Space',
    'context': 'ippc2023',
    'tags': '',
    'viz': 'UAVsViz.UAVsVisualizer',
    'generator': 'UAVInstanceGenerator.UAVInstanceGenerator'
}