info = {
    'name': 'Wildfire_POMDP',
    'description': 'A boolean version of the wildfire fighting domain. This is the pomdp version',
    'context': 'ippc2014',
    'tags': '',
    'viz': 'None'
}
