info = {
    'name': 'CartPole_Discrete',
    'description': 'A simple continuous state discrete action MDP for the classical cart-pole system by Rich Sutton, with actions that describe the direction of the force applied to the cart.',
    'context': 'gym',
    'tags': '',
    'viz': 'CartPoleViz.CartPoleVisualizer'
}