info = {
    'name': 'Epidemic',
    'description': 'SIR model with nonlinear infection dynamics, age groups, hospitalization and multiple interventions.',
    'context': '',
    'tags': '',
    'viz': 'None'
}