info = {
    'name': 'MountainCar_Discrete',
    'description': 'A simple continuous MDP with discrete actions for the classical mountain car control problem.',
    'context': 'gym',
    'tags': '',
    'viz': 'MountainCarViz.MountainCarVisualizer'
}