info = {
    'name': 'Tamarisk_POMDP',
    'description': 'The agent manages the spread of an invasive plant species, by manually intervening to eridaticate them or restore the native species. This is the pomdp version',
    'context': 'ippc2014',
    'tags': '',
    'viz': 'None'
}