info = {
    'name': 'Sokoban',
    'description': 'A person pushes boxes in a warehouse onto designated storage areas, difficult domain due to dead ends.',
    'context': 'arcade',
    'tags': '',
    'viz': 'None'
}