info = {
    'name': 'Quadcopter',
    'description': 'Control a swarm of four-propeller drones in 3D space.',
    'context': 'physics',
    'tags': '',
    'viz': 'QuadcopterViz.QuadcopterVisualizer'
}