info = {
    'name': 'RecSim',
    'description': 'A problem of recommendation systems, with consumers and providers',
    'context': 'ippc2023',
    'tags': '',
    'viz': 'RecSimViz.RecSimVisualizer',
    'generator': 'RecSimInstanceGenerator.RecSimInstanceGenerator'
}