info = {
    'name': 'RaceCar',
    'description': 'A simple continuous MDP for the racecar problem',
    'context': 'ippc2023',
    'tags': '',
    'viz': 'RacecarViz.RacecarVisualizer',
    'generator': 'RaceCarInstanceGenerator.RaceCarInstanceGenerator'
}